"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from flexprice_py.types import UnrecognizedStr
from typing import Literal, Union


AlertState = Union[
    Literal[
        "ok",
        "info",
        "warning",
        "in_alarm",
    ],
    UnrecognizedStr,
]



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from flexprice_py.types import (
    BaseModel,
    dto_updateinvoicerequest as types_dto_updateinvoicerequest,
)
from flexprice_py.utils import FieldMetadata, PathParamMetadata, RequestMetadata
from typing_extensions import Annotated, TypedDict


class UpdateInvoiceRequestTypedDict(TypedDict):
    id: str
    r"""Invoice ID"""
    body: types_dto_updateinvoicerequest.DtoUpdateInvoiceRequestTypedDict
    r"""Invoice Update Request"""


class UpdateInvoiceRequest(BaseModel):
    id: Annotated[
        str, FieldMetadata(path=PathParamMetadata(style="simple", explode=False))
    ]
    r"""Invoice ID"""

    body: Annotated[
        types_dto_updateinvoicerequest.DtoUpdateInvoiceRequest,
        FieldMetadata(request=RequestMetadata(media_type="application/json")),
    ]
    r"""Invoice Update Request"""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from flexprice_py.types import BaseModel, UNSET_SENTINEL
from pydantic import model_serializer
from typing import Optional
from typing_extensions import NotRequired, TypedDict


class DtoTriggerForceRunResponseTypedDict(TypedDict):
    end_time: NotRequired[str]
    message: NotRequired[str]
    mode: NotRequired[str]
    r"""\"custom\" or \"automatic\" """
    start_time: NotRequired[str]
    workflow_id: NotRequired[str]


class DtoTriggerForceRunResponse(BaseModel):
    end_time: Optional[str] = None

    message: Optional[str] = None

    mode: Optional[str] = None
    r"""\"custom\" or \"automatic\" """

    start_time: Optional[str] = None

    workflow_id: Optional[str] = None

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(
            ["end_time", "message", "mode", "start_time", "workflow_id"]
        )
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k)

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from flexprice_py.types import UnrecognizedStr
from typing import Literal, Union


ApplicationStatus = Union[
    Literal[
        "applied",
        "failed",
        "pending",
        "skipped",
        "cancelled",
    ],
    UnrecognizedStr,
]



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from .basesdk import BaseSDK
from flexprice_py import dtos, errors, types, utils
from flexprice_py._hooks import HookContext
from flexprice_py.types import OptionalNullable, UNSET
from flexprice_py.utils.unmarshal_json_response import unmarshal_json_response
from typing import Any, Dict, List, Mapping, Optional, Union


class Features(BaseSDK):
    def create_feature(
        self,
        *,
        name: str,
        type_: types.FeatureType,
        alert_settings: Optional[
            Union[types.AlertSettings, types.AlertSettingsTypedDict]
        ] = None,
        description: Optional[str] = None,
        lookup_key: Optional[str] = None,
        metadata: Optional[Dict[str, str]] = None,
        meter: Optional[
            Union[types.DtoCreateMeterRequest, types.DtoCreateMeterRequestTypedDict]
        ] = None,
        meter_id: Optional[str] = None,
        unit_plural: Optional[str] = None,
        unit_singular: Optional[str] = None,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> types.DtoFeatureResponse:
        r"""Create feature

        Use when defining a new feature or capability to gate or meter (e.g. feature flags or usage-based limits). Ideal for boolean or usage features.

        :param name:
        :param type:
        :param alert_settings:
        :param description:
        :param lookup_key:
        :param metadata:
        :param meter:
        :param meter_id:
        :param unit_plural:
        :param unit_singular:
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = types.DtoCreateFeatureRequest(
            alert_settings=utils.get_pydantic_model(
                alert_settings, Optional[types.AlertSettings]
            ),
            description=description,
            lookup_key=lookup_key,
            metadata=metadata,
            meter=utils.get_pydantic_model(
                meter, Optional[types.DtoCreateMeterRequest]
            ),
            meter_id=meter_id,
            name=name,
            type=type_,
            unit_plural=unit_plural,
            unit_singular=unit_singular,
        )

        req = self._build_request(
            method="POST",
            path="/features",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=True,
            request_has_path_params=False,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            get_serialized_body=lambda: utils.serialize_request_body(
                request, False, False, "json", types.DtoCreateFeatureRequest
            ),
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = self.do_request(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="createFeature",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "201", "application/json"):
            return unmarshal_json_response(types.DtoFeatureResponse, http_res)
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                errors.ErrorsErrorResponseData, http_res
            )
            raise errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                errors.ErrorsErrorResponseData, http_res
            )
            raise errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise errors.FlexpriceDefaultError("Unexpected response received", http_res)

    async def create_feature_async(
        self,
        *,
        name: str,
        type_: types.FeatureType,
        alert_settings: Optional[
            Union[types.AlertSettings, types.AlertSettingsTypedDict]
        ] = None,
        description: Optional[str] = None,
        lookup_key: Optional[str] = None,
        metadata: Optional[Dict[str, str]] = None,
        meter: Optional[
            Union[types.DtoCreateMeterRequest, types.DtoCreateMeterRequestTypedDict]
        ] = None,
        meter_id: Optional[str] = None,
        unit_plural: Optional[str] = None,
        unit_singular: Optional[str] = None,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> types.DtoFeatureResponse:
        r"""Create feature

        Use when defining a new feature or capability to gate or meter (e.g. feature flags or usage-based limits). Ideal for boolean or usage features.

        :param name:
        :param type:
        :param alert_settings:
        :param description:
        :param lookup_key:
        :param metadata:
        :param meter:
        :param meter_id:
        :param unit_plural:
        :param unit_singular:
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = types.DtoCreateFeatureRequest(
            alert_settings=utils.get_pydantic_model(
                alert_settings, Optional[types.AlertSettings]
            ),
            description=description,
            lookup_key=lookup_key,
            metadata=metadata,
            meter=utils.get_pydantic_model(
                meter, Optional[types.DtoCreateMeterRequest]
            ),
            meter_id=meter_id,
            name=name,
            type=type_,
            unit_plural=unit_plural,
            unit_singular=unit_singular,
        )

        req = self._build_request_async(
            method="POST",
            path="/features",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=True,
            request_has_path_params=False,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            get_serialized_body=lambda: utils.serialize_request_body(
                request, False, False, "json", types.DtoCreateFeatureRequest
            ),
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = await self.do_request_async(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="createFeature",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "201", "application/json"):
            return unmarshal_json_response(types.DtoFeatureResponse, http_res)
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                errors.ErrorsErrorResponseData, http_res
            )
            raise errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                errors.ErrorsErrorResponseData, http_res
            )
            raise errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise errors.FlexpriceDefaultError("Unexpected response received", http_res)

    def query_feature(
        self,
        *,
        end_time: Optional[str] = None,
        expand: Optional[str] = None,
        feature_ids: Optional[List[str]] = None,
        filters: Optional[
            Union[List[types.FilterCondition], List[types.FilterConditionTypedDict]]
        ] = None,
        limit: Optional[int] = None,
        lookup_key: Optional[str] = None,
        lookup_keys: Optional[List[str]] = None,
        meter_ids: Optional[List[str]] = None,
        name_contains: Optional[str] = None,
        offset: Optional[int] = None,
        order: Optional[types.FeatureFilterOrder] = None,
        sort: Optional[
            Union[List[types.SortCondition], List[types.SortConditionTypedDict]]
        ] = None,
        start_time: Optional[str] = None,
        status: Optional[types.Status] = None,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> types.DtoListFeaturesResponse:
        r"""Query features

        Use when listing or searching features (e.g. catalog or entitlement setup). Returns a paginated list; supports filtering and sorting.

        :param end_time:
        :param expand:
        :param feature_ids: Feature specific filters
        :param filters: filters allows complex filtering based on multiple fields
        :param limit:
        :param lookup_key:
        :param lookup_keys:
        :param meter_ids:
        :param name_contains:
        :param offset:
        :param order:
        :param sort:
        :param start_time:
        :param status:
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = types.FeatureFilter(
            end_time=end_time,
            expand=expand,
            feature_ids=feature_ids,
            filters=utils.get_pydantic_model(
                filters, Optional[List[types.FilterCondition]]
            ),
            limit=limit,
            lookup_key=lookup_key,
            lookup_keys=lookup_keys,
            meter_ids=meter_ids,
            name_contains=name_contains,
            offset=offset,
            order=order,
            sort=utils.get_pydantic_model(sort, Optional[List[types.SortCondition]]),
            start_time=start_time,
            status=status,
        )

        req = self._build_request(
            method="POST",
            path="/features/search",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=True,
            request_has_path_params=False,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            get_serialized_body=lambda: utils.serialize_request_body(
                request, False, False, "json", types.FeatureFilter
            ),
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = self.do_request(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="queryFeature",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return unmarshal_json_response(types.DtoListFeaturesResponse, http_res)
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                errors.ErrorsErrorResponseData, http_res
            )
            raise errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                errors.ErrorsErrorResponseData, http_res
            )
            raise errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise errors.FlexpriceDefaultError("Unexpected response received", http_res)

    async def query_feature_async(
        self,
        *,
        end_time: Optional[str] = None,
        expand: Optional[str] = None,
        feature_ids: Optional[List[str]] = None,
        filters: Optional[
            Union[List[types.FilterCondition], List[types.FilterConditionTypedDict]]
        ] = None,
        limit: Optional[int] = None,
        lookup_key: Optional[str] = None,
        lookup_keys: Optional[List[str]] = None,
        meter_ids: Optional[List[str]] = None,
        name_contains: Optional[str] = None,
        offset: Optional[int] = None,
        order: Optional[types.FeatureFilterOrder] = None,
        sort: Optional[
            Union[List[types.SortCondition], List[types.SortConditionTypedDict]]
        ] = None,
        start_time: Optional[str] = None,
        status: Optional[types.Status] = None,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> types.DtoListFeaturesResponse:
        r"""Query features

        Use when listing or searching features (e.g. catalog or entitlement setup). Returns a paginated list; supports filtering and sorting.

        :param end_time:
        :param expand:
        :param feature_ids: Feature specific filters
        :param filters: filters allows complex filtering based on multiple fields
        :param limit:
        :param lookup_key:
        :param lookup_keys:
        :param meter_ids:
        :param name_contains:
        :param offset:
        :param order:
        :param sort:
        :param start_time:
        :param status:
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = types.FeatureFilter(
            end_time=end_time,
            expand=expand,
            feature_ids=feature_ids,
            filters=utils.get_pydantic_model(
                filters, Optional[List[types.FilterCondition]]
            ),
            limit=limit,
            lookup_key=lookup_key,
            lookup_keys=lookup_keys,
            meter_ids=meter_ids,
            name_contains=name_contains,
            offset=offset,
            order=order,
            sort=utils.get_pydantic_model(sort, Optional[List[types.SortCondition]]),
            start_time=start_time,
            status=status,
        )

        req = self._build_request_async(
            method="POST",
            path="/features/search",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=True,
            request_has_path_params=False,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            get_serialized_body=lambda: utils.serialize_request_body(
                request, False, False, "json", types.FeatureFilter
            ),
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = await self.do_request_async(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="queryFeature",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return unmarshal_json_response(types.DtoListFeaturesResponse, http_res)
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                errors.ErrorsErrorResponseData, http_res
            )
            raise errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                errors.ErrorsErrorResponseData, http_res
            )
            raise errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise errors.FlexpriceDefaultError("Unexpected response received", http_res)

    def update_feature(
        self,
        *,
        id: str,
        alert_settings: Optional[
            Union[types.AlertSettings, types.AlertSettingsTypedDict]
        ] = None,
        description: Optional[str] = None,
        filters: Optional[
            Union[List[types.MeterFilter], List[types.MeterFilterTypedDict]]
        ] = None,
        metadata: Optional[Dict[str, str]] = None,
        name: Optional[str] = None,
        unit_plural: Optional[str] = None,
        unit_singular: Optional[str] = None,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> types.DtoFeatureResponse:
        r"""Update feature

        Use when changing feature definition (e.g. name, type, or meter). Request body contains the fields to update.

        :param id: Feature ID
        :param alert_settings:
        :param description:
        :param filters:
        :param metadata:
        :param name:
        :param unit_plural:
        :param unit_singular:
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = dtos.UpdateFeatureRequest(
            id=id,
            body=types.DtoUpdateFeatureRequest(
                alert_settings=utils.get_pydantic_model(
                    alert_settings, Optional[types.AlertSettings]
                ),
                description=description,
                filters=utils.get_pydantic_model(
                    filters, Optional[List[types.MeterFilter]]
                ),
                metadata=metadata,
                name=name,
                unit_plural=unit_plural,
                unit_singular=unit_singular,
            ),
        )

        req = self._build_request(
            method="PUT",
            path="/features/{id}",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=True,
            request_has_path_params=True,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            get_serialized_body=lambda: utils.serialize_request_body(
                request.body, False, False, "json", types.DtoUpdateFeatureRequest
            ),
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = self.do_request(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="updateFeature",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "404", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return unmarshal_json_response(types.DtoFeatureResponse, http_res)
        if utils.match_response(http_res, ["400", "404"], "application/json"):
            response_data = unmarshal_json_response(
                errors.ErrorsErrorResponseData, http_res
            )
            raise errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                errors.ErrorsErrorResponseData, http_res
            )
            raise errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise errors.FlexpriceDefaultError("Unexpected response received", http_res)

    async def update_feature_async(
        self,
        *,
        id: str,
        alert_settings: Optional[
            Union[types.AlertSettings, types.AlertSettingsTypedDict]
        ] = None,
        description: Optional[str] = None,
        filters: Optional[
            Union[List[types.MeterFilter], List[types.MeterFilterTypedDict]]
        ] = None,
        metadata: Optional[Dict[str, str]] = None,
        name: Optional[str] = None,
        unit_plural: Optional[str] = None,
        unit_singular: Optional[str] = None,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> types.DtoFeatureResponse:
        r"""Update feature

        Use when changing feature definition (e.g. name, type, or meter). Request body contains the fields to update.

        :param id: Feature ID
        :param alert_settings:
        :param description:
        :param filters:
        :param metadata:
        :param name:
        :param unit_plural:
        :param unit_singular:
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = dtos.UpdateFeatureRequest(
            id=id,
            body=types.DtoUpdateFeatureRequest(
                alert_settings=utils.get_pydantic_model(
                    alert_settings, Optional[types.AlertSettings]
                ),
                description=description,
                filters=utils.get_pydantic_model(
                    filters, Optional[List[types.MeterFilter]]
                ),
                metadata=metadata,
                name=name,
                unit_plural=unit_plural,
                unit_singular=unit_singular,
            ),
        )

        req = self._build_request_async(
            method="PUT",
            path="/features/{id}",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=True,
            request_has_path_params=True,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            get_serialized_body=lambda: utils.serialize_request_body(
                request.body, False, False, "json", types.DtoUpdateFeatureRequest
            ),
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = await self.do_request_async(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="updateFeature",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "404", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return unmarshal_json_response(types.DtoFeatureResponse, http_res)
        if utils.match_response(http_res, ["400", "404"], "application/json"):
            response_data = unmarshal_json_response(
                errors.ErrorsErrorResponseData, http_res
            )
            raise errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                errors.ErrorsErrorResponseData, http_res
            )
            raise errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise errors.FlexpriceDefaultError("Unexpected response received", http_res)

    def delete_feature(
        self,
        *,
        id: str,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> types.DtoSuccessResponse:
        r"""Delete feature

        Use when retiring a feature (e.g. deprecated capability). Returns 200 with success message.

        :param id: Feature ID
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = dtos.DeleteFeatureRequest(
            id=id,
        )

        req = self._build_request(
            method="DELETE",
            path="/features/{id}",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=False,
            request_has_path_params=True,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = self.do_request(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="deleteFeature",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "404", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return unmarshal_json_response(types.DtoSuccessResponse, http_res)
        if utils.match_response(http_res, ["400", "404"], "application/json"):
            response_data = unmarshal_json_response(
                errors.ErrorsErrorResponseData, http_res
            )
            raise errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                errors.ErrorsErrorResponseData, http_res
            )
            raise errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise errors.FlexpriceDefaultError("Unexpected response received", http_res)

    async def delete_feature_async(
        self,
        *,
        id: str,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> types.DtoSuccessResponse:
        r"""Delete feature

        Use when retiring a feature (e.g. deprecated capability). Returns 200 with success message.

        :param id: Feature ID
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = dtos.DeleteFeatureRequest(
            id=id,
        )

        req = self._build_request_async(
            method="DELETE",
            path="/features/{id}",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=False,
            request_has_path_params=True,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = await self.do_request_async(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="deleteFeature",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "404", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return unmarshal_json_response(types.DtoSuccessResponse, http_res)
        if utils.match_response(http_res, ["400", "404"], "application/json"):
            response_data = unmarshal_json_response(
                errors.ErrorsErrorResponseData, http_res
            )
            raise errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                errors.ErrorsErrorResponseData, http_res
            )
            raise errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise errors.FlexpriceDefaultError("Unexpected response received", http_res)



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .alertsettings import AlertSettings, AlertSettingsTypedDict
from .featuretype import FeatureType
from .status import Status
from flexprice_py.types import BaseModel, UNSET_SENTINEL
from pydantic import model_serializer
from typing import Dict, Optional
from typing_extensions import NotRequired, TypedDict


class GithubComFlexpriceFlexpriceInternalDomainFeatureFeatureTypedDict(TypedDict):
    alert_settings: NotRequired[AlertSettingsTypedDict]
    created_at: NotRequired[str]
    created_by: NotRequired[str]
    description: NotRequired[str]
    environment_id: NotRequired[str]
    id: NotRequired[str]
    lookup_key: NotRequired[str]
    metadata: NotRequired[Dict[str, str]]
    meter_id: NotRequired[str]
    name: NotRequired[str]
    status: NotRequired[Status]
    tenant_id: NotRequired[str]
    type: NotRequired[FeatureType]
    unit_plural: NotRequired[str]
    unit_singular: NotRequired[str]
    updated_at: NotRequired[str]
    updated_by: NotRequired[str]


class GithubComFlexpriceFlexpriceInternalDomainFeatureFeature(BaseModel):
    alert_settings: Optional[AlertSettings] = None

    created_at: Optional[str] = None

    created_by: Optional[str] = None

    description: Optional[str] = None

    environment_id: Optional[str] = None

    id: Optional[str] = None

    lookup_key: Optional[str] = None

    metadata: Optional[Dict[str, str]] = None

    meter_id: Optional[str] = None

    name: Optional[str] = None

    status: Optional[Status] = None

    tenant_id: Optional[str] = None

    type: Optional[FeatureType] = None

    unit_plural: Optional[str] = None

    unit_singular: Optional[str] = None

    updated_at: Optional[str] = None

    updated_by: Optional[str] = None

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(
            [
                "alert_settings",
                "created_at",
                "created_by",
                "description",
                "environment_id",
                "id",
                "lookup_key",
                "metadata",
                "meter_id",
                "name",
                "status",
                "tenant_id",
                "type",
                "unit_plural",
                "unit_singular",
                "updated_at",
                "updated_by",
            ]
        )
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k)

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from flexprice_py.types import BaseModel, UNSET_SENTINEL
from pydantic import model_serializer
from typing import Dict, Optional
from typing_extensions import NotRequired, TypedDict


class DtoUpdateInvoiceRequestTypedDict(TypedDict):
    due_date: NotRequired[str]
    invoice_pdf_url: NotRequired[str]
    r"""invoice_pdf_url is the URL where customers can download the PDF version of this invoice"""
    metadata: NotRequired[Dict[str, str]]


class DtoUpdateInvoiceRequest(BaseModel):
    due_date: Optional[str] = None

    invoice_pdf_url: Optional[str] = None
    r"""invoice_pdf_url is the URL where customers can download the PDF version of this invoice"""

    metadata: Optional[Dict[str, str]] = None

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(["due_date", "invoice_pdf_url", "metadata"])
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k)

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from flexprice_py.types import BaseModel, UNSET_SENTINEL
from pydantic import model_serializer
from typing import Dict, Optional
from typing_extensions import NotRequired, TypedDict


class DtoCreateCreditNoteLineItemRequestTypedDict(TypedDict):
    amount: str
    r"""amount is the monetary amount to be credited for this line item"""
    invoice_line_item_id: str
    r"""invoice_line_item_id is the unique identifier of the invoice line item being credited"""
    display_name: NotRequired[str]
    r"""display_name is an optional human-readable name for this credit note line item"""
    metadata: NotRequired[Dict[str, str]]


class DtoCreateCreditNoteLineItemRequest(BaseModel):
    amount: str
    r"""amount is the monetary amount to be credited for this line item"""

    invoice_line_item_id: str
    r"""invoice_line_item_id is the unique identifier of the invoice line item being credited"""

    display_name: Optional[str] = None
    r"""display_name is an optional human-readable name for this credit note line item"""

    metadata: Optional[Dict[str, str]] = None

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(["display_name", "metadata"])
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k)

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m
